#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import functools
from abc import ABC, abstractmethod
from typing import Any, List, Mapping, Optional, Protocol, Tuple

from airbyte_cdk.sources.connector_state_manager import ConnectorStateManager
from airbyte_cdk.sources.message import MessageRepository
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.streams.concurrent.partitions.record import Record
from airbyte_cdk.sources.streams.concurrent.state_converter import ConcurrentStreamStateConverter


def _extract_value(mapping: Mapping[str, Any], path: List[str]) -> Any:
    return functools.reduce(lambda a, b: a[b], path, mapping)


class Comparable(Protocol):
    """Protocol for annotating comparable types."""

    @abstractmethod
    def __lt__(self: "Comparable", other: "Comparable") -> bool:
        pass


class CursorField:
    def __init__(self, cursor_field_key: str) -> None:
        self._cursor_field_key = cursor_field_key

    def extract_value(self, record: Record) -> Comparable:
        cursor_value = record.data.get(self._cursor_field_key)
        if cursor_value is None:
            raise ValueError(f"Could not find cursor field {self._cursor_field_key} in record")
        return cursor_value  # type: ignore  # we assume that the value the path points at is a comparable


class Cursor(ABC):
    @abstractmethod
    def observe(self, record: Record) -> None:
        """
        Indicate to the cursor that the record has been emitted
        """
        raise NotImplementedError()

    @abstractmethod
    def close_partition(self, partition: Partition) -> None:
        """
        Indicate to the cursor that the partition has been successfully processed
        """
        raise NotImplementedError()


class NoopCursor(Cursor):
    def observe(self, record: Record) -> None:
        pass

    def close_partition(self, partition: Partition) -> None:
        pass


class ConcurrentCursor(Cursor):
    _START_BOUNDARY = 0
    _END_BOUNDARY = 1

    def __init__(
        self,
        stream_name: str,
        stream_namespace: Optional[str],
        stream_state: Any,
        message_repository: MessageRepository,
        connector_state_manager: ConnectorStateManager,
        connector_state_converter: ConcurrentStreamStateConverter,
        cursor_field: CursorField,
        slice_boundary_fields: Optional[Tuple[str, str]],
    ) -> None:
        self._stream_name = stream_name
        self._stream_namespace = stream_namespace
        self._message_repository = message_repository
        self._connector_state_converter = connector_state_converter
        self._connector_state_manager = connector_state_manager
        self._cursor_field = cursor_field
        # To see some example where the slice boundaries might not be defined, check https://github.com/airbytehq/airbyte/blob/1ce84d6396e446e1ac2377362446e3fb94509461/airbyte-integrations/connectors/source-stripe/source_stripe/streams.py#L363-L379
        self._slice_boundary_fields = slice_boundary_fields if slice_boundary_fields else tuple()
        self._most_recent_record: Optional[Record] = None
        self._has_closed_at_least_one_slice = False
        self._state = connector_state_converter.get_concurrent_stream_state(stream_state)

    def observe(self, record: Record) -> None:
        if self._slice_boundary_fields:
            # Given that slicing is done using the cursor field, we don't need to observe the record as we assume slices will describe what
            # has been emitted. Assuming there is a chance that records might not be yet populated for the most recent slice, use a lookback
            # window
            return

        if not self._most_recent_record or self._extract_cursor_value(self._most_recent_record) < self._extract_cursor_value(record):
            self._most_recent_record = record

    def _extract_cursor_value(self, record: Record) -> Comparable:
        return self._cursor_field.extract_value(record)

    def close_partition(self, partition: Partition) -> None:
        slice_count_before = len(self._state["slices"])
        self._add_slice_to_state(partition)
        if slice_count_before < len(self._state["slices"]):
            self._merge_partitions()
            self._emit_state_message()
        self._has_closed_at_least_one_slice = True

    def _add_slice_to_state(self, partition: Partition) -> None:
        if self._slice_boundary_fields:
            if "slices" not in self._state:
                self._state["slices"] = []
            self._state["slices"].append(
                {
                    "start": self._extract_from_slice(partition, self._slice_boundary_fields[self._START_BOUNDARY]),
                    "end": self._extract_from_slice(partition, self._slice_boundary_fields[self._END_BOUNDARY]),
                }
            )
        elif self._most_recent_record:
            if self._has_closed_at_least_one_slice:
                raise ValueError(
                    "Given that slice_boundary_fields is not defined and that per-partition state is not supported, only one slice is "
                    "expected."
                )

            self._state["slices"].append(
                {
                    "start": 0,  # FIXME this only works with int datetime
                    "end": self._extract_cursor_value(self._most_recent_record),
                }
            )

    def _emit_state_message(self) -> None:
        self._connector_state_manager.update_state_for_stream(self._stream_name, self._stream_namespace, self._state)
        state_message = self._connector_state_manager.create_state_message(
            self._stream_name, self._stream_namespace, send_per_stream_state=True
        )
        self._message_repository.emit_message(state_message)

    def _merge_partitions(self) -> None:
        self._state["slices"] = self._connector_state_converter.merge_intervals(self._state["slices"])

    def _extract_from_slice(self, partition: Partition, key: str) -> Comparable:
        try:
            _slice = partition.to_slice()
            if not _slice:
                raise KeyError(f"Could not find key `{key}` in empty slice")
            return _slice[key]  # type: ignore  # we expect the devs to specify a key that would return a Comparable
        except KeyError as exception:
            raise KeyError(f"Partition is expected to have key `{key}` but could not be found") from exception



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import logging
import time
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Any, Callable, List


class ThreadPoolManager:
    """
    Wrapper to abstract away the threadpool and the logic to wait for pending tasks to be completed.
    """

    DEFAULT_SLEEP_TIME = 0.1
    DEFAULT_MAX_QUEUE_SIZE = 10_000

    def __init__(
        self,
        threadpool: ThreadPoolExecutor,
        logger: logging.Logger,
        max_concurrent_tasks: int = DEFAULT_MAX_QUEUE_SIZE,
        sleep_time: float = DEFAULT_SLEEP_TIME,
    ):
        """
        :param threadpool: The threadpool to use
        :param logger: The logger to use
        :param max_concurrent_tasks: The maximum number of tasks that can be pending at the same time
        :param sleep_time: How long to sleep if there are too many pending tasks
        """
        self._threadpool = threadpool
        self._logger = logger
        self._max_concurrent_tasks = max_concurrent_tasks
        self._sleep_time = sleep_time
        self._futures: List[Future[Any]] = []

    def submit(self, function: Callable[..., Any], *args: Any) -> None:
        # Submit a task to the threadpool, waiting if there are too many pending tasks
        self._wait_while_too_many_pending_futures(self._futures)
        self._futures.append(self._threadpool.submit(function, *args))

    def _wait_while_too_many_pending_futures(self, futures: List[Future[Any]]) -> None:
        # Wait until the number of pending tasks is < self._max_concurrent_tasks
        while True:
            self._prune_futures(futures)
            if len(futures) < self._max_concurrent_tasks:
                break
            self._logger.info("Main thread is sleeping because the task queue is full...")
            time.sleep(self._sleep_time)

    def _prune_futures(self, futures: List[Future[Any]]) -> None:
        """
        Take a list in input and remove the futures that are completed. If a future has an exception, it'll raise and kill the stream
        operation.

        Pruning this list safely relies on the assumptions that only the main thread can modify the list of futures.
        """
        if len(futures) < self._max_concurrent_tasks:
            return

        for index in reversed(range(len(futures))):
            future = futures[index]
            optional_exception = future.exception()
            if optional_exception:
                exception = RuntimeError(f"Failed reading with error: {optional_exception}")
                self._stop_and_raise_exception(exception)

            if future.done():
                futures.pop(index)

    def shutdown(self) -> None:
        self._threadpool.shutdown(wait=False, cancel_futures=True)

    def is_done(self) -> bool:
        return all([f.done() for f in self._futures])

    def check_for_errors_and_shutdown(self) -> None:
        """
        Check if any of the futures have an exception, and raise it if so. If all futures are done, shutdown the threadpool.
        If the futures are not done, raise an exception.
        :return:
        """
        exceptions_from_futures = [f for f in [future.exception() for future in self._futures] if f is not None]
        if exceptions_from_futures:
            exception = RuntimeError(f"Failed reading with errors: {exceptions_from_futures}")
            self._stop_and_raise_exception(exception)
        else:
            futures_not_done = [f for f in self._futures if not f.done()]
            if futures_not_done:
                exception = RuntimeError(f"Failed reading with futures not done: {futures_not_done}")
                self._stop_and_raise_exception(exception)
            else:
                self.shutdown()

    def _stop_and_raise_exception(self, exception: BaseException) -> None:
        self.shutdown()
        raise exception



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from queue import Queue

from airbyte_cdk.sources.concurrent_source.partition_generation_completed_sentinel import PartitionGenerationCompletedSentinel
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream
from airbyte_cdk.sources.streams.concurrent.partitions.types import QueueItem


class PartitionEnqueuer:
    """
    Generates partitions from a partition generator and puts them in a queue.
    """

    def __init__(self, queue: Queue[QueueItem]) -> None:
        """
        :param queue:  The queue to put the partitions in.
        :param sentinel: The sentinel to put in the queue when all the partitions have been generated.
        """
        self._queue = queue

    def generate_partitions(self, stream: AbstractStream) -> None:
        """
        Generate partitions from a partition generator and put them in a queue.
        When all the partitions are added to the queue, a sentinel is added to the queue to indicate that all the partitions have been generated.

        If an exception is encountered, the exception will be caught and put in the queue.

        This method is meant to be called in a separate thread.
        :param partition_generator: The partition Generator
        :return:
        """
        try:
            for partition in stream.generate_partitions():
                self._queue.put(partition)
            self._queue.put(PartitionGenerationCompletedSentinel(stream))
        except Exception as e:
            self._queue.put(e)



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

import json
import logging
import os
from typing import Iterator, List
from unittest import TestCase
from unittest.mock import Mock, patch

from airbyte_cdk.sources.abstract_source import AbstractSource
from airbyte_cdk.test.entrypoint_wrapper import read
from airbyte_protocol.models import (
    AirbyteAnalyticsTraceMessage,
    AirbyteErrorTraceMessage,
    AirbyteLogMessage,
    AirbyteMessage,
    AirbyteRecordMessage,
    AirbyteStateMessage,
    AirbyteStreamStatus,
    AirbyteStreamStatusTraceMessage,
    AirbyteTraceMessage,
    ConfiguredAirbyteCatalog,
    Level,
    StreamDescriptor,
    TraceType,
    Type,
)

_A_RECORD = AirbyteMessage(
    type=Type.RECORD,
    record=AirbyteRecordMessage(stream="stream", data={"record key": "record value"}, emitted_at=0)
)
_A_STATE_MESSAGE = AirbyteMessage(
    type=Type.STATE,
    state=AirbyteStateMessage(data={"state key": "state value"},)
)
_A_LOG = AirbyteMessage(
    type=Type.LOG,
    log=AirbyteLogMessage(level=Level.INFO, message="This is an Airbyte log message")
)
_AN_ERROR_MESSAGE = AirbyteMessage(
    type=Type.TRACE,
    trace=AirbyteTraceMessage(
        type=TraceType.ERROR,
        emitted_at=0,
        error=AirbyteErrorTraceMessage(message="AirbyteErrorTraceMessage message"),
    ),
)
_AN_ANALYTIC_MESSAGE = AirbyteMessage(
    type=Type.TRACE,
    trace=AirbyteTraceMessage(
        type=TraceType.ANALYTICS,
        emitted_at=0,
        analytics=AirbyteAnalyticsTraceMessage(type="an analytic type", value="an analytic value"),
    ),
)

_A_STREAM_NAME = "a stream name"
_A_CONFIG = {"config_key": "config_value"}
_A_CATALOG = ConfiguredAirbyteCatalog.parse_obj(
    {
        "streams": [
            {
                "stream": {
                    "name": "a_stream_name",
                    "json_schema": {},
                    "supported_sync_modes": ["full_refresh"],
                },
                "sync_mode": "full_refresh",
                "destination_sync_mode": "append",
            }
        ]
    }
)
_A_STATE = {"state_key": "state_value"}
_A_LOG_MESSAGE = "a log message"


def _a_status_message(stream_name: str, status: AirbyteStreamStatus) -> AirbyteMessage:
    return AirbyteMessage(
        type=Type.TRACE,
        trace=AirbyteTraceMessage(
            type=TraceType.STREAM_STATUS,
            emitted_at=0,
            stream_status=AirbyteStreamStatusTraceMessage(
                stream_descriptor=StreamDescriptor(name=stream_name),
                status=status,
            ),
        ),
    )


def _to_entrypoint_output(messages: List[AirbyteMessage]) -> Iterator[str]:
    return (message.json(exclude_unset=True) for message in messages)


def _a_mocked_source() -> AbstractSource:
    source = Mock(spec=AbstractSource)
    source.message_repository = None
    return source


def _validate_tmp_json_file(expected, file_path) -> None:
    with open(file_path) as file:
        assert json.load(file) == expected


def _validate_tmp_catalog(expected, file_path) -> None:
    assert ConfiguredAirbyteCatalog.parse_file(file_path) == expected


def _create_tmp_file_validation(entrypoint, expected_config, expected_catalog, expected_state):
    def _validate_tmp_files(self):
        _validate_tmp_json_file(expected_config, entrypoint.return_value.parse_args.call_args.args[0][2])
        _validate_tmp_catalog(expected_catalog, entrypoint.return_value.parse_args.call_args.args[0][4])
        _validate_tmp_json_file(expected_state, entrypoint.return_value.parse_args.call_args.args[0][6])
        return entrypoint.return_value.run.return_value
    return _validate_tmp_files


class EntrypointWrapperTest(TestCase):
    def setUp(self) -> None:
        self._a_source = _a_mocked_source()

    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_when_read_then_ensure_parameters(self, entrypoint):
        entrypoint.return_value.run.side_effect = _create_tmp_file_validation(entrypoint, _A_CONFIG, _A_CATALOG, _A_STATE)

        read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)

        entrypoint.assert_called_once_with(self._a_source)
        entrypoint.return_value.run.assert_called_once_with(entrypoint.return_value.parse_args.return_value)
        assert entrypoint.return_value.parse_args.call_count == 1
        assert entrypoint.return_value.parse_args.call_args.args[0][0] == "read"
        assert entrypoint.return_value.parse_args.call_args.args[0][1] == "--config"
        assert entrypoint.return_value.parse_args.call_args.args[0][3] == "--catalog"

    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_when_read_then_ensure_files_are_temporary(self, entrypoint):
        read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)

        assert not os.path.exists(entrypoint.return_value.parse_args.call_args.args[0][2])
        assert not os.path.exists(entrypoint.return_value.parse_args.call_args.args[0][4])
        assert not os.path.exists(entrypoint.return_value.parse_args.call_args.args[0][6])

    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_given_logging_during_run_when_read_then_output_has_logs(self, entrypoint):
        def _do_some_logging(self):
            logging.getLogger("any logger").info(_A_LOG_MESSAGE)
            return entrypoint.return_value.run.return_value
        entrypoint.return_value.run.side_effect = _do_some_logging

        output = read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)

        assert len(output.logs) == 1
        assert output.logs[0].log.message == _A_LOG_MESSAGE

    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_given_record_when_read_then_output_has_record(self, entrypoint):
        entrypoint.return_value.run.return_value = _to_entrypoint_output([_A_RECORD])
        output = read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)
        assert output.records == [_A_RECORD]

    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_given_state_message_when_read_then_output_has_state_message(self, entrypoint):
        entrypoint.return_value.run.return_value = _to_entrypoint_output([_A_STATE_MESSAGE])
        output = read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)
        assert output.state_messages == [_A_STATE_MESSAGE]

    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_given_state_message_and_records_when_read_then_output_has_records_and_state_message(self, entrypoint):
        entrypoint.return_value.run.return_value = _to_entrypoint_output([_A_RECORD, _A_STATE_MESSAGE])
        output = read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)
        assert output.records_and_state_messages == [_A_RECORD, _A_STATE_MESSAGE]

    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_given_log_when_read_then_output_has_log(self, entrypoint):
        entrypoint.return_value.run.return_value = _to_entrypoint_output([_A_LOG])
        output = read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)
        assert output.logs == [_A_LOG]

    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_given_trace_message_when_read_then_output_has_trace_messages(self, entrypoint):
        entrypoint.return_value.run.return_value = _to_entrypoint_output([_AN_ANALYTIC_MESSAGE])
        output = read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)
        assert output.analytics_messages == [_AN_ANALYTIC_MESSAGE]

    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_given_stream_statuses_when_read_then_return_statuses(self, entrypoint):
        status_messages = [
            _a_status_message(_A_STREAM_NAME, AirbyteStreamStatus.STARTED),
            _a_status_message(_A_STREAM_NAME, AirbyteStreamStatus.COMPLETE)
        ]
        entrypoint.return_value.run.return_value = _to_entrypoint_output(status_messages)
        output = read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)
        assert output.get_stream_statuses(_A_STREAM_NAME) == [AirbyteStreamStatus.STARTED, AirbyteStreamStatus.COMPLETE]

    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_given_stream_statuses_for_many_streams_when_read_then_filter_other_streams(self, entrypoint):
        status_messages = [
            _a_status_message(_A_STREAM_NAME, AirbyteStreamStatus.STARTED),
            _a_status_message("another stream name", AirbyteStreamStatus.INCOMPLETE),
            _a_status_message(_A_STREAM_NAME, AirbyteStreamStatus.COMPLETE)
        ]
        entrypoint.return_value.run.return_value = _to_entrypoint_output(status_messages)
        output = read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)
        assert len(output.get_stream_statuses(_A_STREAM_NAME)) == 2

    @patch('airbyte_cdk.test.entrypoint_wrapper.print', create=True)
    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_given_unexpected_exception_when_read_then_print(self, entrypoint, print_mock):
        entrypoint.return_value.run.side_effect = ValueError("This error should be printed")
        read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)
        assert print_mock.call_count > 0

    @patch('airbyte_cdk.test.entrypoint_wrapper.print', create=True)
    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_given_expected_exception_when_read_then_do_not_print(self, entrypoint, print_mock):
        entrypoint.return_value.run.side_effect = ValueError("This error should be printed")
        read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE, expecting_exception=True)
        assert print_mock.call_count == 0

    @patch("airbyte_cdk.test.entrypoint_wrapper.AirbyteEntrypoint")
    def test_given_uncaught_exception_when_read_then_output_has_error(self, entrypoint):
        entrypoint.return_value.run.side_effect = ValueError("This error should be printed")
        output = read(self._a_source, _A_CONFIG, _A_CATALOG, _A_STATE)
        assert output.errors



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import base64
import hashlib
from typing import Any, Optional


def hash(value: Any, hash_type: str = "md5", salt: Optional[str] = None) -> str:
    """
      Implementation of a custom Jinja2 hash filter
      Hash type defaults to 'md5' if one is not specified.

      If you are using this has function for GDPR compliance, then
      you should probably also pass in a salt as discussed in:
      https://security.stackexchange.com/questions/202022/hashing-email-addresses-for-gdpr-compliance

      This can be used in a low code connector definition under the AddFields transformation.
      For example:

    rates_stream:
      $ref: "#/definitions/base_stream"
      $parameters:
        name: "rates"
        primary_key: "date"
        path: "/exchangerates_data/latest"
      transformations:
        - type: AddFields
          fields:
            - path: ["some_new_path"]
              value: "{{ record['rates']['CAD'] | hash('md5', 'mysalt')  }}"



      :param value: value to be hashed
      :param hash_type: valid hash type
      :param salt: a salt that will be combined with the value to ensure that the hash created for a given value on this system
                   is different from the hash created for that value on other systems.
      :return: computed hash as a hexadecimal string
    """
    hash_func = getattr(hashlib, hash_type, None)

    if hash_func:
        hash_obj = hash_func()
        hash_obj.update(str(value).encode("utf-8"))
        if salt:
            hash_obj.update(str(salt).encode("utf-8"))
        computed_hash: str = hash_obj.hexdigest()
    else:
        raise AttributeError("No hashing function named {hname}".format(hname=hash_type))

    return computed_hash


def base64encode(value: str) -> str:
    """
    Implementation of a custom Jinja2 base64encode filter

    For example:

      OAuthAuthenticator:
        $ref: "#/definitions/OAuthAuthenticator"
        $parameters:
          name: "client_id"
          value: "{{ config['client_id'] | base64encode }}"

    :param value: value to be encoded in base64
    :return: base64 encoded string
    """

    return base64.b64encode(value.encode("utf-8")).decode()


def base64decode(value: str) -> str:
    """
    Implementation of a custom Jinja2 base64decode filter

    For example:

      OAuthAuthenticator:
        $ref: "#/definitions/OAuthAuthenticator"
        $parameters:
          name: "client_id"
          value: "{{ config['client_id'] | base64decode }}"

    :param value: value to be decoded from base64
    :return: base64 decoded string
    """

    return base64.b64decode(value.encode("utf-8")).decode()


_filters_list = [hash, base64encode, base64decode]
filters = {f.__name__: f for f in _filters_list}



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, List, MutableMapping, Optional


class ConcurrencyCompatibleStateType(Enum):
    date_range = "date-range"


class ConcurrentStreamStateConverter(ABC):
    START_KEY = "start"
    END_KEY = "end"

    def get_concurrent_stream_state(self, state: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
        if self.is_state_message_compatible(state):
            return state
        return self.convert_from_sequential_state(state)

    @staticmethod
    def is_state_message_compatible(state: MutableMapping[str, Any]) -> bool:
        return state.get("state_type") in [t.value for t in ConcurrencyCompatibleStateType]

    @abstractmethod
    def convert_from_sequential_state(self, stream_state: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
        """
        Convert the state message to the format required by the ThreadBasedConcurrentStream.

        e.g.
        {
            "state_type": ConcurrencyCompatibleStateType.date_range.value,
            "metadata": { … },
            "slices": [
                {starts: 0, end: 1617030403, finished_processing: true}]
        }
        """
        ...

    @abstractmethod
    def convert_to_sequential_state(self, stream_state: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
        """
        Convert the state message from the concurrency-compatible format to the stream's original format.

        e.g.
        { "created": 1617030403 }
        """
        ...

    def _get_latest_complete_time(self, slices: List[MutableMapping[str, Any]]) -> Optional[Any]:
        """
        Get the latest time before which all records have been processed.
        """
        if slices:
            first_interval = self.merge_intervals(slices)[0][self.END_KEY]
            return first_interval
        else:
            return None

    @staticmethod
    @abstractmethod
    def increment(timestamp: Any) -> Any:
        """
        Increment a timestamp by a single unit.
        """
        ...

    @classmethod
    def merge_intervals(cls, intervals: List[MutableMapping[str, Any]]) -> List[MutableMapping[str, Any]]:
        sorted_intervals = sorted(intervals, key=lambda x: (x[cls.START_KEY], x[cls.END_KEY]))
        if len(sorted_intervals) > 0:
            merged_intervals = [sorted_intervals[0]]
        else:
            return []
        for interval in sorted_intervals[1:]:
            if interval[cls.START_KEY] <= cls.increment(merged_intervals[-1][cls.END_KEY]):
                merged_intervals[-1][cls.END_KEY] = interval[cls.END_KEY]
            else:
                merged_intervals.append(interval)

        return merged_intervals


class EpochValueConcurrentStreamStateConverter(ConcurrentStreamStateConverter):
    def __init__(self, cursor_field: str):
        self._cursor_field = cursor_field

    def convert_from_sequential_state(self, stream_state: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
        """
        e.g.
        { "created": 1617030403 }
        =>
        {
            "state_type": "date-range",
            "metadata": { … },
            "slices": [
                {starts: 0, end: 1617030403, finished_processing: true}
            ]
        }
        """
        if self.is_state_message_compatible(stream_state):
            return stream_state
        if self._cursor_field in stream_state:
            slices = [
                {
                    self.START_KEY: 0,
                    self.END_KEY: stream_state[self._cursor_field],
                },
            ]
        else:
            slices = []
        return {
            "state_type": ConcurrencyCompatibleStateType.date_range.value,
            "slices": slices,
            "legacy": stream_state,
        }

    def convert_to_sequential_state(self, stream_state: MutableMapping[str, Any]) -> Any:
        """
        e.g.
        {
            "state_type": "date-range",
            "metadata": { … },
            "slices": [
                {starts: 0, end: 1617030403, finished_processing: true}
            ]
        }
        =>
        { "created": 1617030403 }
        """
        if self.is_state_message_compatible(stream_state):
            legacy_state = stream_state.get("legacy", {})
            if slices := stream_state.pop("slices", None):
                legacy_state.update({self._cursor_field: self._get_latest_complete_time(slices)})
            return legacy_state
        else:
            return stream_state

    @staticmethod
    def increment(timestamp: Any) -> Any:
        return timestamp + 1



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.


class HttpResponse:
    def __init__(self, body: str, status_code: int = 200):
        self._body = body
        self._status_code = status_code

    @property
    def body(self) -> str:
        return self._body

    @property
    def status_code(self) -> int:
        return self._status_code



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#
import concurrent
import logging
from queue import Queue
from typing import Iterable, Iterator, List

from airbyte_cdk.models import AirbyteMessage
from airbyte_cdk.sources.concurrent_source.concurrent_read_processor import ConcurrentReadProcessor
from airbyte_cdk.sources.concurrent_source.partition_generation_completed_sentinel import PartitionGenerationCompletedSentinel
from airbyte_cdk.sources.concurrent_source.thread_pool_manager import ThreadPoolManager
from airbyte_cdk.sources.message import InMemoryMessageRepository, MessageRepository
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream
from airbyte_cdk.sources.streams.concurrent.partition_enqueuer import PartitionEnqueuer
from airbyte_cdk.sources.streams.concurrent.partition_reader import PartitionReader
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.streams.concurrent.partitions.record import Record
from airbyte_cdk.sources.streams.concurrent.partitions.types import PartitionCompleteSentinel, QueueItem
from airbyte_cdk.sources.utils.slice_logger import DebugSliceLogger, SliceLogger


class ConcurrentSource:
    """
    A Source that reads data from multiple AbstractStreams concurrently.
    It does so by submitting partition generation, and partition read tasks to a thread pool.
    The tasks asynchronously add their output to a shared queue.
    The read is done when all partitions for all streams were generated and read.
    """

    DEFAULT_TIMEOUT_SECONDS = 900

    @staticmethod
    def create(
        num_workers: int,
        initial_number_of_partitions_to_generate: int,
        logger: logging.Logger,
        slice_logger: SliceLogger,
        message_repository: MessageRepository,
        timeout_seconds: int = DEFAULT_TIMEOUT_SECONDS,
    ) -> "ConcurrentSource":
        threadpool = ThreadPoolManager(
            concurrent.futures.ThreadPoolExecutor(max_workers=num_workers, thread_name_prefix="workerpool"),
            logger,
        )
        return ConcurrentSource(
            threadpool, logger, slice_logger, message_repository, initial_number_of_partitions_to_generate, timeout_seconds
        )

    def __init__(
        self,
        threadpool: ThreadPoolManager,
        logger: logging.Logger,
        slice_logger: SliceLogger = DebugSliceLogger(),
        message_repository: MessageRepository = InMemoryMessageRepository(),
        initial_number_partitions_to_generate: int = 1,
        timeout_seconds: int = DEFAULT_TIMEOUT_SECONDS,
    ) -> None:
        """
        :param threadpool: The threadpool to submit tasks to
        :param logger: The logger to log to
        :param slice_logger: The slice logger used to create messages on new slices
        :param message_repository: The repository to emit messages to
        :param initial_number_partitions_to_generate: The initial number of concurrent partition generation tasks. Limiting this number ensures will limit the latency of the first records emitted. While the latency is not critical, emitting the records early allows the platform and the destination to process them as early as possible.
        :param timeout_seconds: The maximum number of seconds to wait for a record to be read from the queue. If no record is read within this time, the source will stop reading and return.
        """
        self._threadpool = threadpool
        self._logger = logger
        self._slice_logger = slice_logger
        self._message_repository = message_repository
        self._initial_number_partitions_to_generate = initial_number_partitions_to_generate
        self._timeout_seconds = timeout_seconds

    def read(
        self,
        streams: List[AbstractStream],
    ) -> Iterator[AirbyteMessage]:
        self._logger.info("Starting syncing")
        stream_instances_to_read_from = self._get_streams_to_read_from(streams)

        # Return early if there are no streams to read from
        if not stream_instances_to_read_from:
            return

        queue: Queue[QueueItem] = Queue()
        concurrent_stream_processor = ConcurrentReadProcessor(
            stream_instances_to_read_from,
            PartitionEnqueuer(queue),
            self._threadpool,
            self._logger,
            self._slice_logger,
            self._message_repository,
            PartitionReader(queue),
        )

        # Enqueue initial partition generation tasks
        yield from self._submit_initial_partition_generators(concurrent_stream_processor)

        # Read from the queue until all partitions were generated and read
        yield from self._consume_from_queue(
            queue,
            concurrent_stream_processor,
        )
        self._threadpool.check_for_errors_and_shutdown()
        self._logger.info("Finished syncing")

    def _submit_initial_partition_generators(self, concurrent_stream_processor: ConcurrentReadProcessor) -> Iterable[AirbyteMessage]:
        for _ in range(self._initial_number_partitions_to_generate):
            status_message = concurrent_stream_processor.start_next_partition_generator()
            if status_message:
                yield status_message

    def _consume_from_queue(
        self,
        queue: Queue[QueueItem],
        concurrent_stream_processor: ConcurrentReadProcessor,
    ) -> Iterable[AirbyteMessage]:
        while airbyte_message_or_record_or_exception := queue.get(block=True, timeout=self._timeout_seconds):
            yield from self._handle_item(
                airbyte_message_or_record_or_exception,
                concurrent_stream_processor,
            )
            if concurrent_stream_processor.is_done() and queue.empty():
                # all partitions were generated and processed. we're done here
                break

    def _handle_item(
        self,
        queue_item: QueueItem,
        concurrent_stream_processor: ConcurrentReadProcessor,
    ) -> Iterable[AirbyteMessage]:
        # handle queue item and call the appropriate handler depending on the type of the queue item
        if isinstance(queue_item, Exception):
            yield from concurrent_stream_processor.on_exception(queue_item)

        elif isinstance(queue_item, PartitionGenerationCompletedSentinel):
            yield from concurrent_stream_processor.on_partition_generation_completed(queue_item)

        elif isinstance(queue_item, Partition):
            concurrent_stream_processor.on_partition(queue_item)
        elif isinstance(queue_item, PartitionCompleteSentinel):
            yield from concurrent_stream_processor.on_partition_complete_sentinel(queue_item)
        elif isinstance(queue_item, Record):
            yield from concurrent_stream_processor.on_record(queue_item)
        else:
            raise ValueError(f"Unknown queue item type: {type(queue_item)}")

    def _get_streams_to_read_from(self, streams: List[AbstractStream]) -> List[AbstractStream]:
        """
        Iterate over the configured streams and return a list of streams to read from.
        If a stream is not configured, it will be skipped.
        If a stream is configured but does not exist in the source and self.raise_exception_on_missing_stream is True, an exception will be raised
        If a stream is not available, it will be skipped
        """
        stream_instances_to_read_from = []
        for stream in streams:
            stream_availability = stream.check_availability()
            if not stream_availability.is_available():
                self._logger.warning(f"Skipped syncing stream '{stream.name}' because it was unavailable. {stream_availability.message()}")
                continue
            stream_instances_to_read_from.append(stream)
        return stream_instances_to_read_from



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from typing import Any, MutableMapping

import pytest
from airbyte_cdk.models import (
    AirbyteStateBlob,
    AirbyteStateMessage,
    AirbyteStateType,
    AirbyteStream,
    AirbyteStreamState,
    StreamDescriptor,
    SyncMode,
)
from airbyte_cdk.sources.connector_state_manager import ConnectorStateManager
from airbyte_cdk.sources.streams.concurrent.state_converter import (
    ConcurrencyCompatibleStateType,
    ConcurrentStreamStateConverter,
    EpochValueConcurrentStreamStateConverter,
)


class MockConcurrentConnectorStateConverter(ConcurrentStreamStateConverter):
    def convert_from_sequential_state(self, state: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
        state["state_type"] = ConcurrencyCompatibleStateType.date_range.value
        return state

    def convert_to_sequential_state(self, state: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
        state.pop("state_type")
        return state

    @staticmethod
    def increment(timestamp: Any) -> Any:
        return timestamp + 1


@pytest.mark.parametrize(
    "stream, input_state, expected_output_state",
    [
        pytest.param(
            AirbyteStream(name="stream1", json_schema={}, supported_sync_modes=[SyncMode.incremental]),
            [],
            {"state_type": ConcurrencyCompatibleStateType.date_range.value},
            id="no-input-state",
        ),
        pytest.param(
            AirbyteStream(name="stream1", json_schema={}, supported_sync_modes=[SyncMode.incremental]),
            [
                AirbyteStateMessage(
                    type=AirbyteStateType.STREAM,
                    stream=AirbyteStreamState(
                        stream_descriptor=StreamDescriptor(name="stream1", namespace=None),
                        stream_state=AirbyteStateBlob.parse_obj({"created_at": "2022_05_22"}),
                    ),
                ),
            ],
            {"created_at": "2022_05_22", "state_type": ConcurrencyCompatibleStateType.date_range.value},
            id="incompatible-input-state",
        ),
        pytest.param(
            AirbyteStream(name="stream1", json_schema={}, supported_sync_modes=[SyncMode.incremental]),
            [
                AirbyteStateMessage(
                    type=AirbyteStateType.STREAM,
                    stream=AirbyteStreamState(
                        stream_descriptor=StreamDescriptor(name="stream1", namespace=None),
                        stream_state=AirbyteStateBlob.parse_obj(
                            {
                                "created_at": "2022_05_22",
                                "state_type": ConcurrencyCompatibleStateType.date_range.value,
                            },
                        ),
                    ),
                ),
            ],
            {"created_at": "2022_05_22", "state_type": ConcurrencyCompatibleStateType.date_range.value},
            id="compatible-input-state",
        ),
    ],
)
def test_concurrent_connector_state_manager_get_stream_state(stream, input_state, expected_output_state):
    state_manager = ConnectorStateManager({"stream1": stream}, input_state)
    state_converter = MockConcurrentConnectorStateConverter()
    assert state_converter.get_concurrent_stream_state(state_manager.get_stream_state("stream1", None)) == expected_output_state


@pytest.mark.parametrize(
    "input_state, is_compatible",
    [
        pytest.param(
            {},
            False,
            id="no-input-state-is-not-compatible",
        ),
        pytest.param(
            {
                "created_at": "2022_05_22",
                "state_type": ConcurrencyCompatibleStateType.date_range.value,
            },
            True,
            id="input-state-with-date_range-is-compatible",
        ),
        pytest.param(
            {
                "created_at": "2022_05_22",
                "state_type": "fake",
            },
            False,
            id="input-state-with-fake-state-type-is-not-compatible",
        ),
        pytest.param(
            {
                "created_at": "2022_05_22",
            },
            False,
            id="input-state-without-state_type-is-not-compatible",
        ),
    ],
)
def test_concurrent_stream_state_converter_is_state_message_compatible(input_state, is_compatible):
    assert ConcurrentStreamStateConverter.is_state_message_compatible(input_state) == is_compatible


@pytest.mark.parametrize(
    "input_intervals, expected_merged_intervals",
    [
        pytest.param(
            [],
            [],
            id="no-intervals",
        ),
        pytest.param(
            [{"start": 0, "end": 1}],
            [{"start": 0, "end": 1}],
            id="single-interval",
        ),
        pytest.param(
            [{"start": 0, "end": 1}, {"start": 0, "end": 1}],
            [{"start": 0, "end": 1}],
            id="duplicate-intervals",
        ),
        pytest.param(
            [{"start": 0, "end": 1}, {"start": 0, "end": 2}],
            [{"start": 0, "end": 2}],
            id="overlapping-intervals",
        ),
        pytest.param(
            [{"start": 0, "end": 3}, {"start": 1, "end": 2}],
            [{"start": 0, "end": 3}],
            id="enclosed-intervals",
        ),
        pytest.param(
            [{"start": 1, "end": 2}, {"start": 0, "end": 1}],
            [{"start": 0, "end": 2}],
            id="unordered-intervals",
        ),
        pytest.param(
            [{"start": 0, "end": 1}, {"start": 2, "end": 3}],
            [{"start": 0, "end": 3}],
            id="adjacent-intervals",
        ),
        pytest.param(
            [{"start": 3, "end": 4}, {"start": 0, "end": 1}],
            [{"start": 0, "end": 1}, {"start": 3, "end": 4}],
            id="nonoverlapping-intervals",
        ),
        pytest.param(
            [{"start": 0, "end": 1}, {"start": 2, "end": 3}, {"start": 10, "end": 11}, {"start": 1, "end": 4}],
            [{"start": 0, "end": 4}, {"start": 10, "end": 11}],
            id="overlapping-and-nonoverlapping-intervals",
        ),
    ],
)
def test_concurrent_stream_state_converter_merge_intervals(input_intervals, expected_merged_intervals):
    return MockConcurrentConnectorStateConverter.merge_intervals(input_intervals) == expected_merged_intervals


@pytest.mark.parametrize(
    "stream, sequential_state, expected_output_state",
    [
        pytest.param(
            AirbyteStream(name="stream1", json_schema={}, supported_sync_modes=[SyncMode.incremental]),
            {},
            {"slices": [], "state_type": ConcurrencyCompatibleStateType.date_range.value, "legacy": {}},
            id="empty-input-state",
        ),
        pytest.param(
            AirbyteStream(name="stream1", json_schema={}, supported_sync_modes=[SyncMode.incremental]),
            {"created": 1617030403},
            {
                "state_type": "date-range",
                "slices": [{"start": 0, "end": 1617030403}],
                "legacy": {"created": 1617030403},
            },
            id="with-input-state",
        ),
    ],
)
def test_epoch_state_converter_convert_from_sequential_state(stream, sequential_state, expected_output_state):
    state_manager = EpochValueConcurrentStreamStateConverter("created")
    assert state_manager.convert_from_sequential_state(sequential_state) == expected_output_state


@pytest.mark.parametrize(
    "stream, concurrent_state, expected_output_state",
    [
        pytest.param(
            AirbyteStream(name="stream1", json_schema={}, supported_sync_modes=[SyncMode.incremental]),
            {"state_type": ConcurrencyCompatibleStateType.date_range.value},
            {},
            id="empty-input-state",
        ),
        pytest.param(
            AirbyteStream(name="stream1", json_schema={}, supported_sync_modes=[SyncMode.incremental]),
            {"state_type": "date-range", "slices": [{"start": 0, "end": 1617030403}]},
            {"created": 1617030403},
            id="with-input-state",
        ),
    ],
)
def test_epoch_state_converter_convert_to_sequential_state(stream, concurrent_state, expected_output_state):
    state_manager = EpochValueConcurrentStreamStateConverter("created")
    assert state_manager.convert_to_sequential_state(concurrent_state) == expected_output_state



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
import sys
from types import TracebackType
from typing import Any, Optional

from airbyte_cdk.utils.traced_exception import AirbyteTracedException


def assemble_uncaught_exception(exception_type: type[BaseException], exception_value: BaseException) -> AirbyteTracedException:
    if issubclass(exception_type, AirbyteTracedException):
        return exception_value  # type: ignore  # validated as part of the previous line
    return AirbyteTracedException.from_exception(exception_value)


def init_uncaught_exception_handler(logger: logging.Logger) -> None:
    """
    Handles uncaught exceptions by emitting an AirbyteTraceMessage and making sure they are not
    printed to the console without having secrets removed.
    """

    def hook_fn(exception_type: type[BaseException], exception_value: BaseException, traceback_: Optional[TracebackType]) -> Any:
        # For developer ergonomics, we want to see the stack trace in the logs when we do a ctrl-c
        if issubclass(exception_type, KeyboardInterrupt):
            sys.__excepthook__(exception_type, exception_value, traceback_)
            return

        logger.fatal(exception_value, exc_info=exception_value)

        # emit an AirbyteTraceMessage for any exception that gets to this spot
        traced_exc = assemble_uncaught_exception(exception_type, exception_value)

        traced_exc.emit_message()

    sys.excepthook = hook_fn
